// dllmain.cpp : Defines the entry point for the DLL application.
#include "pch.h"

#include <iostream>
#include <Windows.h>
#include <MinHook.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#include <fstream>
#include <string>
#include <sstream>
#include <iomanip>
#include <cstdio>
#include <functional>
#include <chrono>
#include <future>

//#include "timercpp.h"

// Need to link with Ws2_32.lib
#pragma comment(lib, "ws2_32.lib")

#define ACT_CUT_TREE 0xD4
#define ACT_TREE_CUTTING_STATE 0xD5

const char* LOG_FILE_NAME = "e:\\MyProg\\ArenaOnline3D\\main\\Debug\\log.txt";

template <typename T>
inline MH_STATUS MH_CreateHookEx(LPVOID pTarget, LPVOID pDetour, T** ppOriginal)
{
    return MH_CreateHook(pTarget, pDetour, reinterpret_cast<LPVOID*>(ppOriginal));
}

template <typename T>
inline MH_STATUS MH_CreateHookApiEx(LPCWSTR pszModule, LPCSTR pszProcName, LPVOID pDetour, T** ppOriginal)
{
    return MH_CreateHookApi(pszModule, pszProcName, pDetour, reinterpret_cast<LPVOID*>(ppOriginal));
}

class Timer
{
public:
    template <class callable, class... arguments>
    Timer(int after, bool async, callable&& f, arguments&&... args)
    {
        std::function<typename std::result_of<callable(arguments...)>::type()> task(std::bind(std::forward<callable>(f), std::forward<arguments>(args)...));
        if (async)
        {
            std::thread([after, task]() {
                std::this_thread::sleep_for(std::chrono::milliseconds(after));
                task();
                }).detach();
        }
        else
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(after));
            task();
        }
    }
};

typedef int (WINAPI* SEND)(SOCKET s, const char* buf, int len, int flags);
typedef int (WINAPI* RECV)(SOCKET s, const char* buf, int len, int flags);

//typedef void (__cdecl* TIMER_CALLBACK)();

typedef std::chrono::high_resolution_clock Clock;

typedef struct _ARENA_PACKET_HEADER {
    WORD MagicBytes;
    WORD Reserved;
    WORD Size;
    BYTE Action;    
} ARENA_PACKET_HEADER, *PARENA_PACKET_HEADER;

typedef struct _ARENA_TREE_ID {
    DWORD Id;
} ARENA_TREE_ID, *PARENA_TREE_ID;

typedef struct _ARENA_TREE_CUTTING {
    DWORD Id;
} ARENA_TREE_CUTTING, *PARENA_TREE_CUTTING;

enum class HOOK_STATE {
    FREE = 0,
    WAIT_CUTTING,
    WAIT_FINISH_CUTTING
};

typedef struct _ARENA_PACKET_HOOK {
    DWORD Id1 = 0;
    DWORD Id2 = 0;
    DWORD Current_tree_cutting_id = 0;
    HOOK_STATE state = HOOK_STATE::FREE;

    SOCKET s;
    int flags;

    char Id1_packet [11];
    char Id2_packet [11];    

    void clear() {
        memset(this, 0, sizeof(*this));
    }

} ARENA_PACKET_HOOK, *PARENA_PACKET_HOOK;

#pragma pack(1)
typedef struct _ARENA_TREE_CUTTING_STATE {
    BYTE State;         //1 - cut started, 0 - cut finished
    DWORD Unknown1;     //always == 0x11AC63
    DWORD Id;           //tree Id
    WORD Unknown2;      //always == 2    
} ARENA_TREE_CUTTING_STATE, *PARENA_TREE_CUTTING_STATE;
#pragma pack()

// Pointer for calling
SEND fpSend = NULL;
RECV fpRecv = NULL;

static ARENA_PACKET_HOOK aph;

//measure time
static std::chrono::time_point<std::chrono::steady_clock> start_time;

unsigned int current_timer = 0x12E1;

static char nibble_to_ascii_hex(UINT8 i) {
    if (i < 10) return i + '0';
    if (i < 16) return i - 10 + 'A';
    return '?';
}

static void print_hex_line(char* dest, const char* buf, const int length) {
    int i = 0;
    for (i = 0; i < length; i++) {
        dest[3 * i + 0] = nibble_to_ascii_hex((buf[i] >> 4) & 0xF);
        dest[3 * i + 1] = nibble_to_ascii_hex(buf[i] & 0xF);
        dest[3 * i + 2] = 0x20;
    }
    dest[3 * i] = 0;
}

bool process_send(std::ofstream& outfile, char* buf, int len)
{
    auto p = reinterpret_cast<PARENA_PACKET_HEADER>(buf);
    
    if (p->MagicBytes == 0x293A) {
        if ((p->Action == ACT_CUT_TREE) && (p->Size == sizeof(DWORD))){
            //this is cut tree packet
            auto data = reinterpret_cast<PARENA_TREE_ID>(buf + 7);

            if (aph.state == HOOK_STATE::FREE) {
                if (aph.Id1 == 0) {
                    aph.Id1 = data->Id;
                    memcpy(aph.Id1_packet, buf, len);
                    outfile << "Captured first tree Id: " << std::hex << aph.Id1 << std::endl;
                }
                else {
                    aph.Id2 = data->Id;
                    memcpy(aph.Id2_packet, buf, len);
                    outfile << "Captured second tree Id: " << std::hex << aph.Id2 << std::endl;
                    outfile << "Waiting for start cutting..." << std::endl;

                    aph.state = HOOK_STATE::WAIT_CUTTING;
                }
            }
            /*if (first_tree == 0){
                first_tree = data->Id;
                outfile << "Captured first tree Id: " << std::hex << first_tree << std::endl;
            }
            else {
                if (second_tree == 0) {
                    second_tree = data->Id;
                    outfile << "Captured second tree Id: " << std::hex << second_tree << std::endl;
                    outfile << "Waiting for start cutting..." << std::endl;

                    abuse_start = true;
                }
                /*else {
                    data->Id = second_tree;
                    outfile << "Swapping trees 0x" << std::hex << first_tree << " and " << second_tree << std::endl;
                    outfile << "You are going to cut second tree now.." << std::endl;

                    first_tree = 0;
                    second_tree = 0;
                }
            }*/

            //outfile << "Tree Id: " << std::hex << data->Id << std::endl;
            return true;
        }        
    }

    return false;
}

int WINAPI DetourSend(SOCKET s, char* buf, int len, int flags)
{    
    std::ofstream outfile;
    outfile.open(LOG_FILE_NAME, std::ios_base::app); // append instead of overwrite
    
    if (!process_send(outfile, buf, len)) {

        //char buf_bytes[4000];
        //memset(buf_bytes, 0, sizeof(buf_bytes));
        //print_hex_line(buf_bytes, buf, len);
        //std::string log = std::string(buf_bytes);

        //outfile << log << std::endl;

    };

    outfile.close();

    return fpSend(s, buf, len, flags);
}

void timer_callback() {
    std::ofstream outfile;
    outfile.open(LOG_FILE_NAME, std::ios_base::app); // append instead of overwrite
    
    if (aph.Current_tree_cutting_id == aph.Id1) {
        send(aph.s, aph.Id2_packet, sizeof(aph.Id2_packet), aph.flags);
        //outfile << "0x" << std::hex << aph.Id2 << std::endl;
    }
    else {
        send(aph.s, aph.Id1_packet, sizeof(aph.Id1_packet), aph.flags);
        //outfile << "0x" << std::hex << aph.Id1 << std::endl;
    }

    outfile << "Timer (delay == " << std::hex << current_timer << ") ->> is trying now to cut another tree with Id: " << std::hex;
    if (aph.Current_tree_cutting_id == aph.Id1) {        
        outfile << "0x" << std::hex << aph.Id2 << std::endl;
    }
    else {        
        outfile << "0x" << std::hex << aph.Id1 << std::endl;
    }

    outfile.close();

    //current_timer += 0x1;
};

bool process_recv(std::ofstream& outfile, char* buf, int len, SOCKET s, int flags) {

    auto p = reinterpret_cast<PARENA_PACKET_HEADER>(buf);

    if ((aph.state == HOOK_STATE::WAIT_CUTTING) || ((aph.state == HOOK_STATE::WAIT_FINISH_CUTTING)))
    {
        if ((p->Action == ACT_TREE_CUTTING_STATE) && (p->Size == 0xB)) {
            auto data = reinterpret_cast<PARENA_TREE_CUTTING_STATE>(buf + 7);
            if ((data->Unknown1 == 0x11AC63) && (data->Unknown2 == 2)) {
                if (data->State == 1) {

                    aph.Current_tree_cutting_id = data->Id;

                    Timer later_test1(current_timer, true, &timer_callback);

                    /*timer_cutting.setTimeout([&]() {
                        //cout << "Hey.. After " << 0x137B << " ms. But I will stop the timer!" << endl;

                        outfile << "Timer ->> trying to cut another tree with Id: " << std::hex;
                        if (data->Id == aph.Id1) {
                            send(s, aph.Id2_packet, sizeof(aph.Id2_packet), flags);
                            outfile << "0x" << std::hex << aph.Id2 << std::endl;
                        }
                        else {
                            send(s, aph.Id1_packet, sizeof(aph.Id1_packet), flags);
                            outfile << "0x" << std::hex << aph.Id1 << std::endl;
                        }

                        timer_cutting.stop();
                        }, 3000);*/                    

                    start_time = Clock::now();
                    outfile << "Start cutting tree Id: " << std::hex << data->Id << std::endl;
                    aph.state = HOOK_STATE::WAIT_FINISH_CUTTING;

                    aph.s = s;
                    aph.flags = flags;
                }
                else {
                    auto finish_time = Clock::now();

                    outfile << "Finished cutting tree Id: " << std::hex << data->Id << std::endl;
                    outfile << "Time: 0x" << std::uppercase << std::hex 
                        << std::chrono::duration_cast<std::chrono::milliseconds>(finish_time - start_time).count() << " ms" << std::endl << std::endl;

                    /*outfile << "Trying to cut another tree with Id: " << std::hex;
                    if (data->Id == aph.Id1) {
                        send(s, aph.Id2_packet, sizeof(aph.Id2_packet), flags);
                        outfile << "0x" << std::hex << aph.Id2 << std::endl;
                    }
                    else {
                        send(s, aph.Id1_packet, sizeof(aph.Id1_packet), flags);
                        outfile << "0x" << std::hex << aph.Id1 << std::endl;
                    }*/
                              
                    aph.clear();
                }
            }
        }
    }

    return false;
}

int WINAPI DetourRecv(SOCKET s, char* buf, int len, int flags)
{
    //call recv first
    auto bytes_recv = fpRecv(s, buf, len, flags);

    std::ofstream outfile;
    outfile.open(LOG_FILE_NAME, std::ios_base::app); // append instead of overwrite
    
    if (!process_recv(outfile, buf, bytes_recv, s, flags)) {

    }

    /*if (bytes_recv < 8000/3-1) {
        char buf_bytes[8000];
        memset(buf_bytes, 0, bytes_recv);
        print_hex_line(buf_bytes, buf, bytes_recv);
        std::string log = std::string(buf_bytes);

        outfile << log << std::endl;
    }
    else {
        outfile << "Crazy..." << std::endl;
    }  */ 

    outfile.close();

    return bytes_recv;
}

int main_attach() {
    // Initialize MinHook.
    if (MH_Initialize() != MH_OK)
    {
        return FALSE;
    }

    // or you can use the new helper function like this.
    if (MH_CreateHookApiEx(L"ws2_32", "send", &DetourSend, &fpSend) != MH_OK)
    {
        return FALSE;
    }

    // or you can use the new helper function like this.
    if (MH_CreateHookApiEx(L"ws2_32", "recv", &DetourRecv, &fpRecv) != MH_OK)
    {
        return FALSE;
    }

    // Enable the hook for MessageBoxW.
    if (MH_EnableHook(&send) != MH_OK)
    {
        return FALSE;
    }    

    // Enable the hook for MessageBoxW.
    if (MH_EnableHook(&recv) != MH_OK)
    {
        return FALSE;
    }

    //MessageBoxW(NULL, L"Hooked successfully!", L"MinHook Sample", MB_OK);

    return TRUE;
}

int main_detach() {
    // Disable the hook for MessageBoxW.
    if (MH_DisableHook(&send) != MH_OK)
    {
        return 1;
    }

    // Disable the hook for MessageBoxW.
    if (MH_DisableHook(&recv) != MH_OK)
    {
        return 1;
    }

    // Uninitialize MinHook.
    if (MH_Uninitialize() != MH_OK)
    {
        return 1;
    }

    return (TRUE);
}

BOOL APIENTRY DllMain( HMODULE hModule,
                       DWORD  ul_reason_for_call,
                       LPVOID lpReserved
                     )
{
    switch (ul_reason_for_call)
    {
    case DLL_PROCESS_ATTACH:        
        return (main_attach());
    case DLL_THREAD_ATTACH:        
    case DLL_THREAD_DETACH:        
    case DLL_PROCESS_DETACH:
        //main_detach();
        break;
    }

    return TRUE;
}

